import os
import numpy as np
import torch
import argparse
from skimage import img_as_float32, img_as_ubyte, io, color, filters
import cv2
import rawpy
import polyblur
import flow_utils

from colour_demosaicing import demosaicing_CFA_Bayer_Malvar2004 as malvar

from handheld_super_resolution import process


if __name__ == "__main__":
    device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')

    #### Argparser

    def str2bool(v):
        v = str(v)

        if isinstance(v, bool):
            return v
        if v.lower() in ('yes', 'true', 't', 'y', '1'):
            return True
        elif v.lower() in ('no', 'false', 'f', 'n', '0'):
            return False
        else:
            raise argparse.ArgumentTypeError('Boolean value expected.')

    parser = argparse.ArgumentParser()
    ## Image parameter
    parser.add_argument('--impath', type=str, help='input image')
    parser.add_argument('--outpath', type=str, help='out image')
    parser.add_argument('--scale', type=float, default=2, help='Scaling factor')
    parser.add_argument('--verbose', type=int, default=1, help='Verbose option (0 to 4)')

    ## Robustness
    parser.add_argument('--t', type=float,  default=0.12, help='Threshold for robustness')
    parser.add_argument('--s1', type=float,  default=2, help='Threshold for robustness')
    parser.add_argument('--s2', type=float,  default=12, help='Threshold for robustness')
    parser.add_argument('--Mt', type=float,  default=0.8, help='Threshold for robustness')
    parser.add_argument('--R_on', type=str2bool,  default=False, help='Whether robustness is activated or not')

    parser.add_argument('--R_denoising_on', type=str2bool, default=True, help='Whether or not the robustness based denoising should be applied')


    ## Post Processing
    parser.add_argument('--post_process', type=str2bool,  default=False, help='Whether post processing should be applied or not')
    parser.add_argument('--do_sharpening', type=str2bool,  default=False, help='Whether sharpening should be applied during post processing')
    parser.add_argument('--radius', type=float,  default=3, help='If sharpening is applied, radius of the unsharp mask')
    parser.add_argument('--amount', type=float,  default=1.5, help='If sharpening is applied, amount of the unsharp mask')
    parser.add_argument('--alpha', type=float,  default=2, help='If sharpening is applied, amount of the unsharp mask')
    parser.add_argument('--beta', type=float,  default=4, help='If sharpening is applied, amount of the unsharp mask')
    parser.add_argument('--c', type=float,  default=0.352, help='If sharpening is applied, amount of the unsharp mask')
    parser.add_argument('--b', type=float,  default=0.768, help='If sharpening is applied, amount of the unsharp mask')
    parser.add_argument('--do_tonemapping', type=str2bool,  default=False, help='Whether tonnemaping should be applied during post processing')
    parser.add_argument('--do_gamma', type=str2bool,  default=False, help='Whether gamma curve should be applied during post processing')
    parser.add_argument('--autobright', type=float, default=1.25, help="multiplicative scalar to enhance exposure.")


    ## Merging (advanced)
    parser.add_argument('--kernel_shape', type=str, default='handheld', help='"handheld" or "iso" : Whether to use steerable or isotropic kernels')
    parser.add_argument('--k_detail', type=float, default=None, help='SNR based by default')
    parser.add_argument('--k_denoise', type=float, default=None, help='SNR based by default')
    parser.add_argument('--k_stretch', type=float, default=4)
    parser.add_argument('--k_shrink', type=float, default=2)

    ## Alignment (advanced)
    parser.add_argument('--ICA_iter', type=int, default=3, help='Number of ICA Iterations')


    args = parser.parse_args()

    impaths = sorted(os.listdir(args.impath))
    impaths = list(filter(lambda x: '.'  != x[0], impaths))
    impaths = list(filter(lambda x: '.dng'  == x.lower()[-4:], impaths))
    n_images = len(impaths)

    print('Parameters:')
    print('')
    print('  Upscaling factor:       %1.1f' % args.scale)
    print('')
    if args.scale == 1:
        print('    Demosaicking with %d images.' % n_images)
    else:
        print('    Super-resolution with %d images.' % n_images)
        if args.scale > 2:
            print('    WARNING: Since the optics and the integration on the sensor limit the aliasing, do not expect more details than that obtained at x2 (refer to our paper and the original publication).')
    print('')
    if args.R_on:
        print('  Robustness:       enabled')
        print('  -------------------------')
        print('  t:                      %1.2f' % args.t)
        print('  s1:                     %1.2f' % args.s1)
        print('  s2:                     %1.2f' % args.s2)
        print('  Mt:                     %1.2f' % args.Mt)
        if args.R_denoising_on:
            print('  Robustness denoising:   enabled')
        else:
            print('  Robustness denoising:   disabled')
        print('                            ' )
    else:
        print('  Robustness:      disabled')
        print('                            ' )

    print('  Alignment:')
    print('  -------------------------')
    print('  ICA Iterations:         %d' % args.ICA_iter)
    print('')
    print('  Fusion:')
    print('  -------------------------')
    print('  Kernel shape:           %s' % args.kernel_shape)
    print('  k_stretch:              %1.2f' % args.k_stretch)
    print('  k_shrink:               %1.2f' % args.k_shrink)
    if args.k_detail is not None:
        print('  k_detail:               %1.2f' % args.k_detail)
    else:
        print('  k_detail:               SNR based' )
    if args.k_denoise is not None:
        print('  k_denoise:              %1.2f' % args.k_denoise)
    else:
        print('  k_denoise:              SNR based' )
    print('')


    def postprocess(img):
        ## Gamma
        img = np.clip(img, 0, 1)
        img = np.maximum(img, 1e-8) ** (1./2.2)
        ## S-curve tone mapping
        img = 3 * img ** 2 - 2 * img ** 3
        ## Sharpening
        # img = polyblur.polyblur_deblurring(img, n_iter=3, alpha=args.alpha, beta=args.beta, c=args.c, b=args.b)
        img = filters.unsharp_mask(img, radius=args.radius, amount=args.amount)
        img = np.clip(img, 0, 1)
        return img


    #### Handheld ####
    print('Processing with handheld super-resolution')
    options = {'verbose' : args.verbose}
    params={"scale": args.scale,
            "merging":{"kernel":args.kernel_shape},
            "robustness":{"on":args.R_on},
            "kanade": {"tuning": {"kanadeIter":args.ICA_iter}}
            }


    params['robustness']['tuning'] = {'t' : args.t,
                                    's1' : args.s1,
                                    's2' : args.s2,        
                                    'Mt' : args.Mt,       
                                    }

    if args.k_detail is not None:
        params['merging']['tuning']['k_detail'] = args.k_detail
    if args.k_denoise is not None:
        params['merging']['tuning']['k_denoise'] = args.k_denoise

    params['merging'] = {'tuning' : {'k_stretch' : args.k_stretch,
                                    'k_shrink' : args.k_shrink
                                    }}
    params['accumulated robustness denoiser'] = {'on': args.R_denoising_on}

    params['post processing'] = {'on':args.post_process,
                        'do sharpening' : args.do_sharpening,
                        'do color correction':True,
                        'do tonemapping':args.do_tonemapping,
                        'do gamma' : args.do_gamma,
                        'do devignette' : False,
            
                        'sharpening' : {'radius': args.radius,
                                        'ammount': args.amount}
                        }

    params['debug'] = True # For accumulated robustness


    handheld_output, debug_dict = process(args.impath, options, params)
    handheld_output = np.nan_to_num(handheld_output)
    handheld_output = np.clip(handheld_output, 0, 1)


    ####  Read the reference ###
    print('Demosaicking using DCRAW and upscaling using nearest neighbor interpolation')
    ref_lr = rawpy.imread(os.path.join(args.impath, impaths[0]))
    ref_lr = ref_lr.postprocess(use_camera_wb=True, gamma=(1.0, 1.0), output_bps=16, output_color=rawpy.ColorSpace.raw, no_auto_bright=True)
    ref_lr = img_as_float32(ref_lr)


    #### Autbright the images - 0.5% of pixels will be saturated.
    # autobright = args.autobright

    handheld_output_small = cv2.resize(handheld_output, dsize=None, fx=1./8, fy=1./8, interpolation=cv2.INTER_LINEAR)
    handheld_output_small = np.mean(handheld_output_small, axis=-1) # grayscale
    nbins = 2048
    count, intensity = np.histogram(handheld_output_small.ravel(), bins=nbins)
    cumulative_count = np.cumsum(count)
    quantile = 1 - (0.5 / 100)
    threshold = count.sum() * quantile

    for n in range(nbins):
        if cumulative_count[n] > threshold:
            top_intensity = intensity[n]
            break

    autobright = 1. / top_intensity

    # autobright = args.autobright
    ref_lr *= autobright
    ref_lr = np.clip(ref_lr, 0, 1)
    handheld_output *= autobright
    handheld_output = np.clip(handheld_output, 0, 1)


    #### Postprocess the images.
    handheld_output_corrected = postprocess(handheld_output)
    ref_lr = postprocess(ref_lr)


    #### Upsample the reference.
    dsize = handheld_output.shape[:2][::-1]
    ref_hr = cv2.resize(ref_lr, dsize=dsize, interpolation=cv2.INTER_LINEAR) 


    # define a faster imsave for large png images
    def imsave(fname, rgb_8bit_data):
        return cv2.imwrite(fname,  cv2.cvtColor(rgb_8bit_data, cv2.COLOR_RGB2BGR ))



    #### Save images ####
    imsave(args.outpath, img_as_ubyte(handheld_output_corrected))
    imsave(args.outpath.replace('output', 'input_0'), img_as_ubyte(ref_hr))

    # Robustness map
    if args.R_on:
        r_acc = debug_dict['accumulated robustness']
        try:
            r_acc = r_acc / (n_images-1)
        except:
            r_acc = r_acc.copy_to_host() / (n_images-1)
        r_acc = cv2.resize(r_acc, dsize=dsize, interpolation=cv2.INTER_NEAREST)
    else:
        r_acc = np.ones_like(handheld_output)
    imsave(args.outpath.replace('output', 'robustness_map'), img_as_ubyte(r_acc))


    print('done')

